#include <iostream>

#include <l4/libkproxy/factory_svr>
#include <l4/re/util/object_registry>
#include <l4/re/util/cap_alloc>
#include <l4/re/util/meta>
#include <l4/cxx/ipc_server>


class Proxy: public L4::Server_object,
             public L4kproxy::Factory_svr,
             public L4kproxy::Factory_interface,
             public L4kproxy::Cap_allocator_interface
{
public:
  Proxy();

  int dispatch(l4_umword_t object, L4::Ipc::Iostream& iostream);

private:
  // L4kproxy::Factory_svr
  virtual L4::Cap<L4::Thread> received_thread(L4::Ipc::Snd_fpage const&);

  // L4kproxy::Factory_interface
  virtual int create_task(L4::Cap<L4::Task> const&, l4_fpage_t const&);
  virtual int create_thread(L4::Cap<L4::Thread> const&);
  virtual int create_factory(L4::Cap<L4::Factory> const&, unsigned long);
  virtual int create_gate(L4::Cap<L4::Kobject> const&, L4::Cap<L4::Thread> const&, l4_umword_t);
  virtual int create_irq(L4::Cap<L4::Irq>const&);
  virtual int create_vm(L4::Cap<L4::Vm>const&);

  // L4kproxy::Cap_allocator_interface
  virtual L4::Cap<void> cap_alloc();
  virtual void cap_free(L4::Cap<void>);
};


Proxy::Proxy()
  : Factory_svr(this, this)
{
}

int Proxy::dispatch(l4_umword_t object, L4::Ipc::Iostream& iostream)
{
  l4_msgtag_t tag;
  iostream >> tag;

  switch (tag.label())
  {
  case L4::Meta::Protocol:
    return L4Re::Util::handle_meta_request<L4::Factory>(iostream);

  case L4::Factory::Protocol:
    return Factory_svr::factory_dispatch(object, iostream);

  default:
    std::cerr << "Proxy: Unsupported protocol: " << tag.label() << std::endl;
    return -L4_EBADPROTO;
  };
}

L4::Cap<L4::Thread> Proxy::received_thread(L4::Ipc::Snd_fpage const& fpage)
{
  (void)fpage;

  // don't care in this example, only used for IPC gates
  return L4::Cap<L4::Thread>::Invalid;
}

int Proxy::create_factory(L4::Cap<L4::Factory> const& target_cap, unsigned long limit)
{
  return l4_error(L4Re::Env::env()->factory()->create_factory(target_cap, limit));
}

int Proxy::create_task(L4::Cap<L4::Task> const& target_cap, l4_fpage_t const& utcb_area)
{
  return l4_error(L4Re::Env::env()->factory()->create_task(target_cap, utcb_area));
}

int Proxy::create_thread(L4::Cap<L4::Thread> const& target_cap)
{
  return l4_error(L4Re::Env::env()->factory()->create_thread(target_cap));
}

int Proxy::create_gate(L4::Cap<L4::Kobject> const& target_cap,
                           L4::Cap<L4::Thread> const& thread_cap,
                           l4_umword_t label)
{
  return l4_error(L4Re::Env::env()->factory()->create_gate(target_cap, thread_cap, label));
}

int Proxy::create_irq(L4::Cap<L4::Irq>const& target_cap)
{
  return l4_error(L4Re::Env::env()->factory()->create_irq(target_cap));
}

int Proxy::create_vm(L4::Cap<L4::Vm>const& target_cap)
{
  return l4_error(L4Re::Env::env()->factory()->create_vm(target_cap));
}

L4::Cap<void> Proxy::cap_alloc()
{
  return L4Re::Util::cap_alloc.alloc<void>();
}

void Proxy::cap_free(L4::Cap<void> capability)
{
  L4Re::Util::cap_alloc.free(capability);
}


int main()
{
  static L4Re::Util::Object_registry registry;
  static L4::Server<> server(l4_utcb());
  static Proxy proxy;

  if (!registry.register_obj(&proxy, "numa_mgr").is_valid())
  {
    std::cerr << "Factory registration failed" << std::endl;
    return 1;
  }

  std::cout << "Ready. Waiting." << std::endl;
  server.loop(registry);
  return 0;
}
